from EEETools.Tools.Other.fernet_handler import FernetHandler
from EEETools.Tools.Other.handler import Handler
from PyQt5.QtWidgets import QWidget
from EEETools import costants
from abc import ABC
import abc, os


class AbstractCostCorrelation(ABC):

    def __init__(self):
        self.name = ""
        self.note = ""
        self.component_type = ""

        self.fernet_handler = FernetHandler()

    def import_from_file(self, file_path):
        """DON'T OVERRIDE THIS METHOD"""
        data = self.fernet_handler.read_file(file_path)
        self.import_xml_tree(data)

    def export_to_file(self, file_path = ""):
        """DON'T OVERRIDE THIS METHOD"""

        if file_path == "":

            correlation_handler = CostCorrelationHandler()
            file_path = correlation_handler.get_file_path(self)

        data = self.export_xml_tree()
        self.fernet_handler.save_file(file_path, data)

    @abc.abstractmethod
    def import_xml_tree(self, data):
        raise NotImplementedError()

    @abc.abstractmethod
    def export_xml_tree(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_cost(self):
        raise NotImplementedError()

    @property
    def currentCEPCI(self):
        return self.__current_cepci

    @currentCEPCI.setter
    def currentCEPCI(self, input_cepci):
        self.__current_cepci = input_cepci

    @property
    @abc.abstractmethod
    def parameter_dict(self) -> dict:
        raise NotImplementedError()

    @abc.abstractmethod
    def set_parameters(self, input_dict):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def coefficient_dict(self):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def coefficients(self):
        raise NotImplementedError()

    @coefficients.setter
    @abc.abstractmethod
    def coefficients(self, input_dict: dict):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def description(self):
        pass

    @property
    @abc.abstractmethod
    def edit_correlation_widget(self) -> QWidget:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def parameter_settings_widget(self):
        raise NotImplementedError()


class CostCorrelationHandler(Handler):

    def __init__(self):

        super().__init__()

        self.data_folder = os.path.join(costants.ROOT_DIR, "res", "Cost Correlation Data")
        self.subclass_directory_path = "EEETools.Tools.CostCorrelations.CorrelationClasses"

        from EEETools.Tools.modules_handler import ModulesHandler
        modules_handler = ModulesHandler()
        self.name_list = modules_handler.name_list

    def check_data_folder(self):

        if not os.path.isdir(self.data_folder):

            try:

                os.mkdir(self.data_folder)

            except:

                pass

        for name in self.name_list:

            folder_name = self.get_module_name(name)
            name_folder_path = os.path.join(self.data_folder, folder_name)

            if not os.path.isdir(name_folder_path):

                try:

                    os.mkdir(name_folder_path)

                except:

                    pass

    def open_file(self, file_path):

        fernet_handler = FernetHandler()
        data = fernet_handler.read_file(file_path)

        __header = data.find("header")
        __correlation_name = __header.get("correlation_class")
        __subclass = self.import_correct_sub_class(__correlation_name)

        return __subclass(data)

    def save_file(self, correlation_class, file_path = ""):

        if file_path == "":

            file_path = self.get_file_path(correlation_class)

        correlation_class.export_to_file(file_path)

    def get_file_path(self, correlation_class):

        file_folder = os.path.join(self.data_folder, self.get_module_name(correlation_class.component_type))

        if os.path.isdir(file_folder):

            return os.path.join(file_folder, self.get_module_name(correlation_class.name) + ".dat")

        else:

            raise FileNotFoundError

    @staticmethod
    def rename_file(new_name, file_path):

        fernet_handler = FernetHandler()
        data = fernet_handler.read_file(file_path)

        header = data.find("header")
        header.attrib["name"] = new_name

        fernet_handler.save_file(file_path, data)


# <----------------------------------------->
# <------------ SUPPORT CLASSES ------------>
# <----------------------------------------->

class Parameter:

    def __init__(self, input_name):

        self.name = input_name
        self.note = ""
        self.measure_unit = ""

        self.value = 0.
        self.__range = Range()

    @property
    def range(self):
        return self.__range

    @range.setter
    def range(self, input_range):
        self.__range.initialize(input_range)


class Range:

    def __init__(self):

        self.low = 0.
        self.high = 0.

    def __str__(self):

        if self.low == self.high:

            if self.low == 0:

                return ""

            else:

                return str(self.low)

        else:

            return str(self.low) + " - " + str(self.high)

    def initialize(self, input_range):

        if type(input_range) is str:

            input_range = input_range.strip()

            if input_range == "":

                self.low = 0.
                self.high = 0.

            elif "-" not in input_range:

                self.low = float(input_range)
                self.high = float(input_range)

            else:

                elements = input_range.split("-")
                self.low = float(elements[0])
                self.high = float(elements[1])

        elif type(input_range) is list:

            self.low = input_range[0]
            self.high = input_range[1]

        elif type(input_range) is Range:

            self.low = input_range.low
            self.high = input_range.high



from EEETools.MainModules.support_blocks import Drawer
from EEETools.MainModules.main_module import Block
import xml.etree.ElementTree as ETree
from EEETools import costants


class Condenser(Block):

    def __init__(self, inputID, main_class):

        super().__init__(inputID, main_class)

        self.type = "condenser"

        if self.main_class.options.condenser_is_dissipative:

            self.has_support_block = True
            self.support_block.append(Drawer(main_class, self, is_input=True))

        else:

            self.has_support_block = False

    def is_ready_for_calculation(self):
        return len(self.input_connections) >= 1 and len(self.output_connections) >= 1

    def prepare_for_calculation(self):

        if self.main_class.options.condenser_is_dissipative:

            self.support_block[0].prepare_for_calculation()

        if self.main_class.options.loss_cost_is_zero:

            new_conn = self.main_class.append_connection(from_block=self)
            new_conn.name = "condenser exergy loss"
            new_conn.automatically_generated_connection = True
            new_conn.exergy_value = self.exergy_balance
            new_conn.is_fluid_stream = False

    def append_excel_connection_list(self, input_list):

        new_input_conn = self.main_class.find_connection_by_index(input_list[0])
        new_output_conn = self.main_class.find_connection_by_index(input_list[1])

        if self.main_class.options.condenser_is_dissipative:

            self.add_connection(new_input_conn, is_input=True, append_to_support_block=0)
            self.add_connection(new_output_conn, is_input=False, append_to_support_block=0)

        else:

            self.add_connection(new_input_conn, is_input=True)
            self.add_connection(new_output_conn, is_input=False)

    def export_xml_other_parameters(self) -> ETree.Element:

        other_tree = ETree.Element("Other")
        return other_tree

    def append_xml_other_parameters(self, input_list: ETree.Element):

        pass

    def export_xml_connection_list(self) -> ETree.Element:

        xml_connection_list = ETree.Element("Connections")

        fluid_connections = ETree.SubElement(xml_connection_list, "FluidConnections")

        if self.main_class.options.condenser_is_dissipative:

            input_connections = self.support_block[0].external_input_connections
            output_connections = self.support_block[0].external_output_connections

        else:

            input_connections = self.external_input_connections
            output_connections = self.external_output_connections

        for input_connection in input_connections:

            if not input_connection.automatically_generated_connection:

                input_xml = ETree.SubElement(fluid_connections, "input")
                input_xml.set("index", str(input_connection.index))

        for output_connection in output_connections:

            if not output_connection.automatically_generated_connection:

                output_xml = ETree.SubElement(fluid_connections, "output")
                output_xml.set("index", str(output_connection.index))

        return xml_connection_list

    def append_xml_connection_list(self, input_list: ETree.Element):

        fluid_connections = input_list.find("FluidConnections")

        if self.main_class.options.condenser_is_dissipative:

            self.__add_connection_by_index(fluid_connections, "input", append_to_support_block=0)
            self.__add_connection_by_index(fluid_connections, "output", append_to_support_block=0)

        else:

            self.__add_connection_by_index(fluid_connections, "input")
            self.__add_connection_by_index(fluid_connections, "output")

    def __add_connection_by_index(self, input_list: ETree.Element, connection_name, append_to_support_block=None):

        if connection_name == "input":

            is_input = True

        else:

            is_input = False

        for connection in input_list.findall(connection_name):

            new_conn = self.main_class.find_connection_by_index(float(connection.get("index")))

            if new_conn is not None:
                self.add_connection(new_conn, is_input, append_to_support_block=append_to_support_block)

    @classmethod
    def return_EES_needed_index(cls):

        return_dict = {"flow input": [1, False],
                       "flow output": [2, False]}

        return return_dict

    @classmethod
    def return_EES_base_equations(cls):

        return_element = dict()

        variables_list = [{"variable": "flow input", "type": costants.ZONE_TYPE_FLOW_RATE},
                          {"variable": "flow output", "type": costants.ZONE_TYPE_FLOW_RATE}]

        return_element.update({"mass_continuity": {"variables": variables_list, "related_option": "none"}})

        variables_list = [{"variable": "flow input", "type": costants.ZONE_TYPE_PRESSURE},
                          {"variable": "flow output", "type": costants.ZONE_TYPE_PRESSURE}]

        return_element.update({"pressure_continuity": {"variables": variables_list, "related_option": "none"}})

        return return_element

    def return_other_zone_connections(self, zone_type, input_connection):

        if zone_type == costants.ZONE_TYPE_FLOW_RATE:

            # In the condenser flow rate is preserved, hence if "input_connection" stream is connected to the condenser
            # block the methods must returns each fluid stream connected to that block

            if self.connection_is_in_connections_list(input_connection):

                return self.get_fluid_stream_connections()

            else:

                return list()

        elif zone_type == costants.ZONE_TYPE_FLUID:

            # In the condenser fluid type is preserved, hence if "input_connection" stream is connected to the condenser
            # block the methods must returns each fluid stream connected to that block

            if self.connection_is_in_connections_list(input_connection):

                return self.get_fluid_stream_connections()

            else:

                return list()

        elif zone_type == costants.ZONE_TYPE_PRESSURE:

            # In the condenser pressure is preserved, hence if "input_connection" stream is connected to the condenser
            # block the methods must returns each fluid stream connected to that block

            if self.connection_is_in_connections_list(input_connection):

                return self.get_fluid_stream_connections()

            else:

                return list()

        else:

            return list()


from EEETools.Tools.EESCodeGenerator.EES_code import Code
from EEETools.Tools.modules_handler import ModulesHandler
from EEETools.Tools.GUIElements.gui_base_classes import *
from PyQt5.QtWidgets import *
from PyQt5.QtGui import QFont, QKeySequence
from PyQt5.QtCore import Qt
from shutil import copy2 as copy_file
import os
from EEETools import costants


class DefineEESTextWidget(QDialog):

    # noinspection PyArgumentList
    def __init__(self, flags=None, *args, **kwargs):

        super().__init__(flags, *args, **kwargs)
        self.modules_handler = ModulesHandler()

        self.init_title()
        self.tables = dict()

        self.main_layout = QVBoxLayout(self)

        self.init_h_layout(0)
        self.init_h_layout(1)
        self.init_h_layout(2)
        self.init_h_layout(3)

        self.setLayout(self.main_layout)

        self.file_manager = None
        self.__enable_buttons()

    def init_title(self, EES_code_name="New EES code"):

        self.EES_code_name = EES_code_name
        self.is_saved = False
        self.__update_title()

    def init_h_layout(self, layout_layer):

        h_layout_new = QHBoxLayout()

        if layout_layer == 0:

            name_label = QLabel("Insert a name:")
            name_label.setFont(QFont("Helvetica 20"))
            name_label.setMinimumWidth(100)
            name_label.setMaximumWidth(100)

            self.name_edit_text = QLineEdit()
            self.name_edit_text.setFont(QFont("Helvetica 20 Bold"))
            self.name_edit_text.textChanged.connect(self.on_text_changed)

            h_layout_new.addWidget(name_label)
            h_layout_new.addWidget(self.name_edit_text)

        elif layout_layer == 1:

            name_label = QLabel("Select the Type:")
            name_label.setFont(QFont("Helvetica 20"))
            name_label.setMinimumWidth(100)
            name_label.setMaximumWidth(100)

            self.type_combobox = QComboBox()
            self.type_combobox.addItems(self.modules_handler.name_list)
            self.type_combobox.currentIndexChanged.connect(self.on_combo_box_changed)
            self.type_combobox.setFont(QFont("Helvetica 20 Bold"))

            h_layout_new.addWidget(name_label)
            h_layout_new.addWidget(self.type_combobox)

        elif layout_layer == 2:

            h_layout_new.addLayout(self.init_left_layout())
            h_layout_new.addLayout(self.init_right_layout())

        else:

            self.button_save = QPushButton()
            self.button_save.setText("Save (CTRL-S)")
            self.button_save.clicked.connect(self.on_save_button_pressed)

            self.button_new = QPushButton()
            self.button_new.setText("New (CTRL-N)")
            self.button_new.clicked.connect(self.on_new_button_pressed)

            self.__set_buttons_shortcut()

            h_layout_new.addWidget(self.button_save)
            h_layout_new.addWidget(self.button_new)

            self.__enable_buttons()

        self.main_layout.addLayout(h_layout_new)

    def init_left_layout(self):

        font = QFont()
        font.setFamily(costants.EES_CODE_FONT_FAMILY)
        font.setBold(True)

        self.input_script = QTextEdit("\"insert here your EES code ...\"")
        self.input_script.setMinimumWidth(400)
        self.input_script.setAcceptRichText(True)
        self.input_script.textChanged.connect(self.on_text_changed)
        self.input_script.setFont(font)

        self.code = Code(self)
        self.on_format_pressed()

        self.button_format = QPushButton()
        self.button_format.setText("Format Code (CTRL-F)")
        self.button_format.pressed.connect(self.on_format_pressed)

        v_layout_left = QVBoxLayout()
        v_layout_left.addWidget(self.input_script)
        v_layout_left.addWidget(self.button_format)

        return v_layout_left

    def init_right_layout(self):

        self.tables = dict()

        v_layout_right = QVBoxLayout()

        v_splitter = QSplitter(Qt.Vertical)
        v_splitter = self.init_table_view(v_splitter, "Input Indices")
        v_splitter = self.init_table_view(v_splitter, "Parameters Indices")
        v_splitter = self.init_table_view(v_splitter, "Equations")

        v_layout_right.addWidget(v_splitter)

        return v_layout_right

    def init_table_view(self, layout, title_str):

        widget = QWidget()
        v_layout = QVBoxLayout()
        title_label_input = QLabel(title_str, font=QFont('Helvetica 30 bold'))

        new_table = AbstractTable(self, title_str)
        new_table.setMinimumWidth(350)
        new_table.setMaximumWidth(350)

        header = new_table.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.Stretch)
        header.setSectionResizeMode(1, QHeaderView.ResizeToContents)

        if title_str in ["Input Indices", "Parameters Indices"]:
            header.setSectionResizeMode(2, QHeaderView.ResizeToContents)

        v_layout.addWidget(title_label_input)
        v_layout.addWidget(new_table)

        widget.setLayout(v_layout)
        layout.addWidget(widget)
        self.tables.update({title_str: new_table})

        return layout

    def open_file(self, file_path):

        try:
            root = self.code.read_file(file_path)

        except:

            save_warning_dialog = WarningDialog(self)

            save_warning_dialog.set_label_text("The file has been compromised and cannot be opened.\n"
                                               "An empty file will be opened instead")

            save_warning_dialog.show_only_one_button = True
            save_warning_dialog.set_button_text("ok", set_button_yes=True)
            save_warning_dialog.show()

            self.__reset_dialog()

        else:

            header = root.find("header")
            code_name = header.get("name")
            component_type = header.get("component_type")

            EES_code = root.find("EESCode").text

            self.name_edit_text.setText(code_name)
            self.input_script.setText(EES_code)
            self.set_combobox_name(component_type)

            file_name = code_name + ".xml"
            folder_name = self.modules_handler.get_module_name(component_type)
            self.EES_code_name = os.path.join(folder_name, file_name)

            self.__update_title()
            self.on_format_pressed()

    def set_combobox_name(self, name):

        index = self.modules_handler.get_name_index(name)

        if index == -1:
            index = 0

        self.type_combobox.setCurrentIndex(index)

    def on_combo_box_changed(self):

        new_model = IndexSetterTableModel(self, "Input Indices")
        self.tables["Input Indices"].setModel(new_model)

        new_model = IndexSetterTableModel(self, "Equations")
        self.tables["Equations"].setModel(new_model)

    def on_text_changed(self):

        try:

            self.__enable_buttons()
            self.is_saved = False
            self.__update_title()

        except:

            pass

    def on_new_button_pressed(self):

        self.new_warning_dialog = WarningDialog(self)
        self.new_warning_dialog.set_label_text("By clicking \"Continue\" all your progress will be lost")
        self.new_warning_dialog.set_buttons_on_click(self.on_new_warning_clicked)
        self.new_warning_dialog.set_button_text("Continue", set_button_yes=True)
        self.new_warning_dialog.set_button_text("Cancel", set_button_yes=False)

        self.new_warning_dialog.show()

    def on_new_warning_clicked(self):

        sender = self.sender()
        if sender.is_yes_button:
            self.__reset_dialog()

        self.new_warning_dialog.close()

    def on_save_button_pressed(self):

        self.save_warning_dialog = WarningDialog(self)
        self.save_warning_dialog.set_label_text("Do you want to save the file? Data will be overwritten")
        self.save_warning_dialog.set_buttons_on_click(self.on_save_warning_clicked)
        self.save_warning_dialog.set_button_text("Continue", set_button_yes=True)
        self.save_warning_dialog.set_button_text("Cancel", set_button_yes=False)

        self.save_warning_dialog.show()

    def on_save_warning_clicked(self):

        sender = self.sender()
        if sender.is_yes_button:
            self.__save()

        self.save_warning_dialog.close()

    def on_format_pressed(self):

        self.code.update()
        self.input_script.setHtml(self.code.rich_text)
        self.input_script.repaint()

        if hasattr(self, 'tables'):
            if "Parameters Indices" in self.tables.keys():
                new_model = IndexSetterTableModel(self, "Parameters Indices")
                new_model.load_data(self.code.inputs)
                self.tables["Parameters Indices"].setModel(new_model)
                self.tables["Parameters Indices"].repaint()

    def __enable_buttons(self):

        self.button_save.setEnabled(self.is_ready_for_saving)

    def __reset_dialog(self):

        self.init_title()
        self.name_edit_text.setText("")
        self.input_script.setText("\"insert here your EES code ...\"")

        self.name_edit_text.setEnabled(True)
        self.type_combobox.setEnabled(True)

        self.on_format_pressed()

    def __save(self):

        return_list = self.code.save()

        folder_name = return_list[0]
        file_name = return_list[1]

        self.is_saved = True
        self.EES_code_name = os.path.join(folder_name, file_name)
        self.name_edit_text.setEnabled(False)
        self.type_combobox.setEnabled(False)
        self.__update_title()

    def __set_buttons_shortcut(self):

        sequence_save = QKeySequence(Qt.CTRL + Qt.Key_S)
        self.button_save.setShortcut(sequence_save)

        sequence_new = QKeySequence(Qt.CTRL + Qt.Key_N)
        self.button_new.setShortcut(sequence_new)

        sequence_format = QKeySequence(Qt.CTRL + Qt.Key_F)
        self.button_format.setShortcut(sequence_format)

    def __update_title(self):

        title = self.EES_code_name + " - "

        if not self.is_saved:
            title += "Not "

        title += "Saved"

        self.setWindowTitle(title)

    @property
    def is_ready_for_saving(self):

        __name = self.name_edit_text.text()
        __input_script = self.input_script.toPlainText()

        return not __name == "" and not (__input_script == "" or __input_script == "\"insert here your EES code ...\"")


class EESTextFilesManager(AbstractFilesManager):

    def __init__(self, flags=None, *args, **kwargs):

        super().__init__(flags, *args, **kwargs)

        self.title = "EES File Manager"
        self.modules_handler = ModulesHandler()

    def on_button_add_pressed(self):

        self.file_edit = DefineEESTextWidget(self)

        if not os.path.isdir(self.selected_element_path):

            dir_name = os.path.dirname(self.selected_element_path)

        else:

            dir_name = self.selected_element_path

        self.file_edit.set_combobox_name(os.path.basename(dir_name))
        self.file_edit.type_combobox.setEnabled(False)
        self.file_edit.show()

    def on_button_modify_pressed(self):

        self.file_edit = DefineEESTextWidget(self)

        self.file_edit.open_file(self.selected_element_path)
        self.file_edit.name_edit_text.setEnabled(False)
        self.file_edit.type_combobox.setEnabled(False)
        self.file_edit.show()

    def on_button_delete_pressed(self):

        dir_name = os.path.dirname(self.selected_element_path)

        dir_name = os.path.basename(dir_name)
        filename = os.path.basename(self.selected_element_path)

        warning_text = "Do you really want to delete file "
        warning_text += os.path.join(dir_name, filename)
        warning_text += "?"

        self.delete_warning_dialog = WarningDialog(self)
        self.delete_warning_dialog.set_label_text(warning_text)
        self.delete_warning_dialog.set_buttons_on_click(self.on_delete_warning_clicked)
        self.delete_warning_dialog.set_button_text("Yes", set_button_yes=True)
        self.delete_warning_dialog.set_button_text("No", set_button_yes=False)

        self.delete_warning_dialog.show()

    def on_delete_warning_clicked(self):

        sender = self.sender()

        if sender.is_yes_button:
            if os.path.exists(self.selected_element_path):
                os.remove(self.selected_element_path)

        self.__enable_buttons()
        self.delete_warning_dialog.close()

    def on_button_rename_pressed(self):

        dir_name = os.path.dirname(self.selected_element_path)

        dir_name = os.path.basename(dir_name)
        filename = os.path.basename(self.selected_element_path)

        label_text = "insert new name for file "
        label_text += os.path.join(dir_name, filename)

        self.rename_line_edit = QLineEdit()
        hint_label = QLabel("new name:")

        hLayout = QHBoxLayout()
        hLayout.addWidget(hint_label)
        hLayout.addWidget(self.rename_line_edit)

        self.rename_warning_dialog = WarningDialog(self)
        self.rename_warning_dialog.set_layout(hLayout)
        self.rename_warning_dialog.set_label_text(label_text)
        self.rename_warning_dialog.set_buttons_on_click(self.on_rename_warning_clicked)
        self.rename_warning_dialog.set_button_text("Rename", set_button_yes=True)
        self.rename_warning_dialog.set_button_text("Cancel", set_button_yes=False)

        self.rename_warning_dialog.show()

    def on_rename_warning_clicked(self):

        sender = self.sender()

        if sender.is_yes_button:
            new_name = self.rename_line_edit.text()
            self.__rename_file(new_name, self.selected_element_path)
            self.__enable_buttons()

        self.rename_warning_dialog.close()

    def on_button_duplicate_pressed(self):

        dir_name = os.path.dirname(self.selected_element_path)

        dir_name = os.path.basename(dir_name)
        filename = os.path.basename(self.selected_element_path)

        warning_text = "Do you really want to duplicate file "
        warning_text += os.path.join(dir_name, filename)
        warning_text += "?"

        self.duplicate_warning_dialog = WarningDialog(self)
        self.duplicate_warning_dialog.set_label_text(warning_text)
        self.duplicate_warning_dialog.set_buttons_on_click(self.on_duplicate_warning_pressed)
        self.duplicate_warning_dialog.set_button_text("Yes", set_button_yes=True)
        self.duplicate_warning_dialog.set_button_text("No", set_button_yes=False)

        self.duplicate_warning_dialog.show()

    def on_duplicate_warning_pressed(self):

        sender = self.sender()

        if sender.is_yes_button:

            if os.path.exists(self.selected_element_path):

                dir_name = os.path.dirname(self.selected_element_path)
                filename = os.path.basename(self.selected_element_path)

                if ".dat" in filename:
                    filename = filename.split(".dat")[0]
                    file_extension = ".dat"

                elif ".xml" in filename:
                    filename = filename.split(".xml")[0]
                    file_extension = ".xml"

                else:

                    file_extension = ""

                i = 1
                new_path = os.path.join(dir_name, filename + "(1)" + file_extension)

                while os.path.exists(new_path):
                    i += 1
                    new_path = os.path.join(dir_name, filename + "(" + str(i) + ")" + file_extension)

                copy_file(self.selected_element_path, new_path)

        self.duplicate_warning_dialog.close()

    def __rename_file(self, new_name, file_path):

        Code.rename_file(new_name, file_path)
        super(EESTextFilesManager, self).__rename_file(new_name, file_path)


import os

# RUN MODE
RUN_MODE = "administrator"
# RUN_MODE = "standard"

# ROOT DIRECTORIES
ROOT_DIR = os.path.dirname(os.path.abspath(os.path.join(__file__, os.pardir)))
RES_DIR = os.path.join(ROOT_DIR, "res")
TEST_RES_DIR = os.path.join(RES_DIR, "testResources")

# FIREBASE CONFIGURATION DICT
FIREBASE_CONFIG = {

            "apiKey": "AIzaSyBQ4ZKw45o3U06gz9u1pmLd53cImzYGHj8",
            "databaseURL":"https://etapp-serggroup-default-rtdb.europe-west1.firebasedatabase.app",
            "authDomain": "etapp-serggroup.firebaseapp.com",
            "projectId": "etapp-serggroup",
            "storageBucket": "etapp-serggroup.appspot.com",
            "messagingSenderId": "12706429785",
            "appId": "1:12706429785:web:66cd6f6b1c1ed13b9e2a48",
            "measurementId": "G-T2SHVKSR9H"

        }

# EES FORMAT STYLES
STYLES = {"error": {"color": "#ff4d00",
                    "font-weight": "bold",
                    "font-style": "normal",
                    "text-decoration": "none"},

          "comments": {"color": "#8C8C8C",
                       "font-weight": "bold",
                       "font-style": "italic",
                       "text-decoration": "none"},

          "variable": {"color": "#0033B3",
                       "font-weight": "bold",
                       "font-style": "normal",
                       "text-decoration": "none"},

          "repeated_keyword": {"color": "#008080",
                               "font-weight": "bold",
                               "font-style": "normal",
                               "text-decoration": "none"},

          "known_keyword": {"color": "#008080",
                            "font-weight": "bold",
                            "font-style": "normal",
                            "text-decoration": "none"},

          "unknown_function": {"color": "#94558D",
                               "font-weight": "bold",
                               "font-style": "italic",
                               "text-decoration": "underline"},

          "known_function": {"color": "#94558D",
                             "font-weight": "bold",
                             "font-style": "normal",
                             "text-decoration": "none"},

          "default": {"color": "#000000",
                      "font-weight": "bold",
                      "font-style": "normal",
                      "text-decoration": "none"}

          }

EES_CODE_FONT_FAMILY = "Courier New"


def get_html_string(key, text):

    if not key in STYLES.keys():
        key = "default"

    style = STYLES[key]

    __html_text = '<span style="'

    __html_text += "color:" + style["color"] + "; "
    __html_text += "font-weight:" + style["font-weight"] + '; '
    __html_text += "font-style:" + style["font-style"] + '; '
    __html_text += "text-decoration:" + style["text-decoration"] + '"'

    __html_text += '>' + text + '</span>'

    return __html_text


# ZONE TYPES
ZONE_TYPE_FLUID = "fluid"
ZONE_TYPE_FLOW_RATE = "flow rate"
ZONE_TYPE_PRESSURE = "pressure"
ZONE_TYPE_ENERGY = "energy"

ZONES = [ZONE_TYPE_ENERGY, ZONE_TYPE_PRESSURE, ZONE_TYPE_FLUID, ZONE_TYPE_FLOW_RATE]


from datetime import date, datetime
import math
import pandas
import os

from EEETools.MainModules.main_module import ArrayHandler
from EEETools.Tools.Other.fernet_handler import FernetHandler


def calculate_excel(excel_path):

    array_handler = import_excel_input(excel_path)
    array_handler.calculate()
    export_solution_to_excel(excel_path, array_handler)


def calculate_dat(dat_path):

    array_handler = import_dat(dat_path)
    array_handler.calculate()
    write_csv_solution(dat_path, array_handler)


def convert_excel_to_dat(excel_path: str):

    array_handler = import_excel_input(excel_path)

    if ".xlsm" in excel_path:

        dat_path = excel_path.replace(".xlsm", ".dat")

    elif ".xlsx" in excel_path:

        dat_path = excel_path.replace(".xlsx", ".dat")

    else:

        dat_path = excel_path.replace(".xls", ".dat")

    export_dat(dat_path, array_handler)


def import_excel_input(excel_path) -> ArrayHandler:

    array_handler = ArrayHandler()

    # import connections
    excel_connection_data = pandas.read_excel(excel_path, sheet_name="Stream")

    for line in excel_connection_data.values:

        line = line.tolist()
        if not math.isnan(line[0]):
            new_conn = array_handler.append_connection()

            new_conn.index = line[0]
            new_conn.name = str(line[1])
            new_conn.exergy_value = line[2]

    # import blocks
    excel_block_data = pandas.read_excel(excel_path, sheet_name="Componenti")

    for line in excel_block_data.values:

        line = line.tolist()

        if not (math.isnan(line[0]) or type(line[0]) is str):

            if line[0] > 0:

                if "Heat Exchanger" in str(line[2]) or "Scambiatore" in str(line[2]):

                    new_block = array_handler.append_block("Heat Exchanger")
                    excel_connection_list = list()
                    excel_connection_list.append(str(line[2]))
                    excel_connection_list.extend(line[5:-1])

                else:

                    new_block = array_handler.append_block(str(line[2]))
                    excel_connection_list = line[5:-1]

                new_block.index = line[0]
                new_block.name = str(line[1])
                new_block.comp_cost = line[3]

                new_block.append_excel_connection_list(excel_connection_list)

            else:

                array_handler.append_excel_costs_and_useful_output(line[5:-1], line[0] == 0, line[3])

    return array_handler


def export_solution_to_excel(excel_path, array_handler: ArrayHandler):

    result_df = get_result_data_frames(array_handler)

    # generation of time stamps for excel sheet name
    today = date.today()
    now = datetime.now()
    today_str = today.strftime("%d %b")
    now_str = now.strftime("%H.%M")

    with pandas.ExcelWriter(excel_path, mode="a") as writer:

        for key in result_df.keys():

            pandas_df = pandas.DataFrame(data=result_df[key])
            pandas_df.to_excel(writer, sheet_name=(key + " - " + today_str + " - " + now_str))


def export_dat(dat_path, array_handler: ArrayHandler):

    fernet = FernetHandler()
    fernet.save_file(dat_path, array_handler.xml)


def import_dat(dat_path) -> ArrayHandler:

    array_handler = ArrayHandler()
    fernet = FernetHandler()
    root = fernet.read_file(dat_path)
    array_handler.xml = root

    return array_handler


def write_csv_solution(dat_path, array_handler):

    result_df = get_result_data_frames(array_handler)

    # generation of time stamps for excel sheet name
    today = date.today()
    now = datetime.now()
    today_str = today.strftime("%d %b")
    now_str = now.strftime("%H.%M")

    dir_path = os.path.dirname(dat_path)

    for key in result_df.keys():

        csv_path = key + " - " + today_str + " - " + now_str + ".csv"
        csv_path = os.path.join(dir_path, csv_path)

        pandas_df = pandas.DataFrame(data=result_df[key])
        pandas_df.to_csv(path_or_buf= csv_path, sep="\t")


def get_result_data_frames(array_handler: ArrayHandler):

    # Stream Solution Data frame generation
    stream_data = {"Stream": list(),
                   "Name": list(),
                   "Exergy Value [kW]": list(),
                   "Specific Cost [Euro/kJ]": list(),
                   "Specific Cost [Euro/kWh]": list(),
                   "Total Cost [Euro/s]": list()}

    for conn in array_handler.connection_list:

        if not conn.is_internal_stream:

            stream_data["Stream"].append(conn.index)
            stream_data["Name"].append(conn.name)
            stream_data["Exergy Value [kW]"].append(conn.exergy_value)
            stream_data["Specific Cost [Euro/kJ]"].append(conn.rel_cost)
            stream_data["Specific Cost [Euro/kWh]"].append(conn.rel_cost*3600)
            stream_data["Total Cost [Euro/s]"].append(conn.rel_cost * conn.exergy_value)

    # Output Stream Data frame generation
    useful_data = {"Stream": list(),
                   "Name": list(),
                   "Exergy Value [kW]": list(),
                   "Specific Cost [Euro/kJ]": list(),
                   "Specific Cost [Euro/kWh]": list(),
                   "Total Cost [Euro/s]": list()}

    for conn in array_handler.useful_effect_connections:

        useful_data["Stream"].append(conn.index)
        useful_data["Name"].append(conn.name)
        useful_data["Exergy Value [kW]"].append(conn.exergy_value)
        useful_data["Specific Cost [Euro/kJ]"].append(conn.rel_cost)
        useful_data["Specific Cost [Euro/kWh]"].append(conn.rel_cost * 3600)
        useful_data["Total Cost [Euro/s]"].append(conn.rel_cost * conn.exergy_value)

    return {"Stream Out": stream_data,
            "Eff Out": useful_data}
